from typing import Optional

from torch.jit import ScriptModule

from nebullvm.base import ModelParams
from nebullvm.inference_learners.pytorch import (
    PytorchBackendInferenceLearner,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager


class BladeDISCInferenceLearner(PytorchBackendInferenceLearner):
    @classmethod
    def from_torch_model(
        cls,
        model: ScriptModule,
        network_parameters: ModelParams,
        input_tfms: Optional[MultiStageTransformation] = None,
        input_data: DataManager = None,
    ):
        return cls(
            torch_model=model,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
            input_data=input_data,
        )



from abc import ABC, abstractmethod
from typing import Any, Optional, Dict, Callable, Tuple

import yaml

from nebullvm.utils.data import DataManager


class BaseCompressor(ABC):
    def __init__(self, config_file: str = None):
        self._config = self._read_config(config_file)

    @abstractmethod
    def compress(
        self,
        model: Any,
        train_input_data: DataManager,
        eval_input_data: DataManager,
        metric_drop_ths: float,
        metric: Callable,
    ) -> Tuple[Any, Optional[float]]:
        raise NotImplementedError()

    def _read_config(self, config_file: Optional[str]) -> Dict:
        config = self._get_default_config()
        if config_file is not None:
            with open(config_file, "r") as f:
                data = yaml.load(f, Loader=yaml.CLoader)
                config.update(data.get(self.config_key, {}))
        return config

    @staticmethod
    @abstractmethod
    def _get_default_config() -> Dict:
        raise NotImplementedError

    @property
    @abstractmethod
    def config_key(self) -> str:
        raise NotImplementedError()



import logging
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Iterable, Sequence, Union, Dict, Callable, List

import torch.nn

from nebullvm.api.frontend.onnx import extract_info_from_np_data
from nebullvm.api.frontend.tf import extract_info_from_tf_data
from nebullvm.api.frontend.torch import extract_info_from_torch_data
from nebullvm.api.huggingface import convert_hf_model, is_dict_type
from nebullvm.api.utils import QUANTIZATION_METRIC_MAP
from nebullvm.base import (
    DeepLearningFramework,
    ModelParams,
    ModelCompiler,
    OptimizationTime,
)
from nebullvm.converters.converters import CrossConverter
from nebullvm.pipelines.steps import build_pipeline_from_model
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.feedback_collector import FEEDBACK_COLLECTOR
from nebullvm.utils.onnx import get_output_sizes_onnx
from nebullvm.utils.optional_modules import tensorflow as tf
from nebullvm.utils.tf import get_outputs_sizes_tf
from nebullvm.utils.torch import get_outputs_sizes_torch


logging.basicConfig(
    format="%(asctime)s %(message)s", datefmt="%d/%m/%Y %I:%M:%S %p"
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _get_dl_framework(model: Any):
    if isinstance(model, torch.nn.Module):
        return DeepLearningFramework.PYTORCH
    elif isinstance(model, tf.Module) and model is not None:
        return DeepLearningFramework.TENSORFLOW
    elif isinstance(model, str):
        return DeepLearningFramework.NUMPY
    else:
        raise TypeError(f"Model type {type(model)} not supported.")


def _check_input_data(input_data: Union[Iterable, Sequence]):
    try:
        input_data[0]
    except:  # noqa E722
        return False
    else:
        return True


INFO_EXTRACTION_DICT: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: extract_info_from_torch_data,
    DeepLearningFramework.TENSORFLOW: extract_info_from_tf_data,
    DeepLearningFramework.NUMPY: extract_info_from_np_data,
}


OUTPUT_SIZE_COMPUTATION_DICT: Dict[DeepLearningFramework, Callable] = {
    DeepLearningFramework.PYTORCH: get_outputs_sizes_torch,
    DeepLearningFramework.TENSORFLOW: get_outputs_sizes_tf,
    DeepLearningFramework.NUMPY: get_output_sizes_onnx,
}


def _extract_info_from_data(
    model: Any,
    input_data: DataManager,
    dl_framework: DeepLearningFramework,
    dynamic_info: Dict,
):
    batch_size, input_sizes, input_types, dynamic_info = INFO_EXTRACTION_DICT[
        dl_framework
    ](
        model,
        input_data,
        batch_size=None,
        input_sizes=None,
        input_types=None,
        dynamic_axis=dynamic_info,
    )
    model_params = ModelParams(
        batch_size=batch_size,
        input_infos=[
            {"size": size, "dtype": dtype}
            for size, dtype in zip(input_sizes, input_types)
        ],
        output_sizes=OUTPUT_SIZE_COMPUTATION_DICT[dl_framework](
            model, input_data[0][0]
        ),
        dynamic_info=dynamic_info,
    )
    return model_params


def _is_huggingface_data(data_sample: Any) -> bool:
    if is_dict_type(data_sample):
        return True
    elif isinstance(data_sample, str):
        return True
    elif isinstance(data_sample[0], str):
        return True
    return False


def optimize_model(
    model: Any,
    input_data: Union[Iterable, Sequence],
    metric_drop_ths: float = None,
    metric: Union[str, Callable] = None,
    optimization_time: str = "constrained",
    dynamic_info: Dict = None,
    config_file: str = None,
    ignore_compilers: List[str] = None,
    **kwargs,
):
    """Optimize the input model regardless of the framework it was used for
    implementing it. The optimized model given as output will share with the
    input one the same API, i.e. the optimized model will have the same
    interface as the original one.

    Args:
        model (Any): The input model.
        input_data (Iterable or Sequence): Input data to be used for
            optimizing the model. Note that if 'unconstrained' is selected as
            `optimization_time`, it would be beneficial to provide at least 100
            data samples in order to use all the techniques supported by
            Nebullvm. The data can be given in either as sequence (data can be
            accessed by "element", e.g. `data[i]`) or iterable (data needs to
            be accessed with loop, e.g. `for x in data`). PyTorch, TensorFlow
            and Onnx respectively accept input tensor in `torch.Tensor`,
            `tf.Tensor` and `np.ndarray` formats. Note that the each input
            sample must be a tuple containing a tuple as first element, the
            `inputs`, and the `label` as second element. The `inputs` needs to
            be passed as tuple even if a single input is needed by the model
            (in this case the `inputs` tuple will contain just an element).
            HuggingFace models can take as data samples both dictionaries or
            strings. Strings will then be converted in data samples using the
            HuggingFace tokenizer which must be given as input when just a
            list of string is provided as input_data (tokenizers can be passed
            as extra arguments of this function using the keyword `tokenizer`).
        metric_drop_ths (float, optional): Maximum reduction in the
            selected metric accepted. No model with an higher error will be
            accepted, i.e. all optimized model having a larger error respect to
            the original one will be discarded, without even considering their
            possible speed-up. Default: None, i.e. no drop in metric accepted.
        metric (Union[Callable, str], optional): The metric to
            be used for accepting or refusing a precision-reduction
            optimization proposal. If none is given but a `metric_drop_ths` is
            received, the `nebullvm.measure.compute_relative_difference`
            metric will be used as default one. A user-defined metric can
            be passed as function accepting as inputs two tuples of tensors
            (produced by the baseline and the optimized model) and the related
            original labels.
            For more information see
            `nebullvm.measure.compute_relative_difference` and
            `nebullvm.measure.compute_accuracy_drop`. `metric`
            accepts as value also a string containing the metric name. At the
            current stage the supported metrics are `"numeric_precision"` and
            `"accuracy"`. Default: `"numeric_precision"`
        optimization_time (OptimizationTime, optional): The optimization time
            mode. It can be either 'constrained' or 'unconstrained'. For
            'constrained' mode just compilers and precision reduction
            techniques are used (no compression). 'Unconstrained' optimization
            allows the usage of more time consuming techniques as pruning and
            distillation. Note that for using many of the sophisticated
            techniques in the 'unconstrained' optimization, a small fine-tuning
            of the model will be needed. Thus we highly recommend to give as
            input_data at least 100 samples for when selecting 'unconstrained'
            optimization. Default: 'constrained'.
        dynamic_info (Dict, optional): Dictionary containing info about the
            dynamic axis. It should contain as keys both "inputs" and "outputs"
            and as values two lists of dictionaries where each dictionary
            represents the dynamic axis information for an input/output tensor.
            The inner dictionary should have as key an integer, i.e. the
            dynamic axis (considering also the batch size) and as value a
            string giving a "tag" to it, e.g. "batch_size". Default: None
        config_file (str, optional): Configuration file containing the
            parameters needed for defining the CompressionStep in the pipeline.
            Default: None.
        ignore_compilers (List, optional): List containing the compilers to be
            ignored during the OptimizerStep. Default: None.

    Returns:
        InferenceLearner: Optimized version of the input model having the same
            interface, imported by its original framework. For instance a
            Pytorch model, when optimized, will return an InferenceLearner
            object that can be call exactly as a PyTorch model (either
            with `model.forward(input)` and `model(input)`), i.e. it will
            take as input and it will return `torch.Tensor`s.
    """
    dl_framework = _get_dl_framework(model)
    optimization_time = OptimizationTime(optimization_time)
    FEEDBACK_COLLECTOR.start_collection(model, framework=dl_framework)
    if metric_drop_ths is not None and metric_drop_ths <= 0:
        metric_drop_ths = None
    elif metric_drop_ths is not None and metric is None:
        metric = "numeric_precision"
    if isinstance(metric, str):
        metric = QUANTIZATION_METRIC_MAP.get(metric)
    needs_conversion_to_hf = False
    if _is_huggingface_data(input_data[0]):
        (
            model,
            input_data,
            input_names,
            output_structure,
            output_type,
        ) = convert_hf_model(model, input_data, **kwargs)
        needs_conversion_to_hf = True
    if _check_input_data(input_data):
        input_data = DataManager(input_data)
    else:
        input_data = DataManager.from_iterable(input_data)
    model_params = _extract_info_from_data(
        model,
        input_data,
        dl_framework,
        dynamic_info,
    )
    converter = CrossConverter()
    optimized_models = []
    with TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        models = converter.convert(model, model_params, tmp_dir, input_data)

        if ignore_compilers is None:
            ignore_compilers = []
        else:
            ignore_compilers = [
                ModelCompiler(compiler) for compiler in ignore_compilers
            ]
        for model in models:
            input_tfms = MultiStageTransformation([])
            pipeline = build_pipeline_from_model(
                model,
                optimization_time,
                metric_drop_ths,
                config_file,
                logger=logger,
            )
            output_dict = pipeline.run(
                model=model,
                input_data=input_data,
                metric_drop_ths=metric_drop_ths,
                metric=metric,
                output_library=dl_framework,
                model_params=model_params,
                input_tfms=input_tfms,
                ignore_compilers=ignore_compilers,
                optimization_time=optimization_time,
            )
            ignore_compilers = output_dict["ignore_compilers"]
            optimized_models.extend(output_dict["optimized_models"])

    optimized_models.sort(key=lambda x: x[1], reverse=False)
    optimal_model = optimized_models[0][0]
    FEEDBACK_COLLECTOR.send_feedback()
    if optimal_model is None:
        logger.warning(
            "No optimized model has been created. This is likely due to a "
            "bug in Nebullvm. Please open an issue and report in details "
            "your use case."
        )
        return optimal_model
    if needs_conversion_to_hf:
        from nebullvm.api.huggingface import HuggingFaceInferenceLearner

        optimal_model = HuggingFaceInferenceLearner(
            core_inference_learner=optimal_model,
            output_structure=output_structure,
            input_names=input_names,
            output_type=output_type,
        )
    return optimal_model



import copy
from abc import ABC, abstractmethod
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, Any, Callable, Optional, Tuple

import numpy as np
import torch.nn
import yaml
from torch.utils.data import DataLoader, Dataset

from nebullvm.compressors.base import BaseCompressor
from nebullvm.utils.data import DataManager
from nebullvm.utils.optional_modules import tensorflow as tf

try:
    from neural_compressor.experimental import Pruning
except ImportError:
    pass


def _get_model_framework(model: Any) -> str:
    if isinstance(model, torch.nn.Module):
        return "torch"
    elif isinstance(model, tf.Module) and model is not None:
        return "tensorflow"
    else:
        return "numpy"


class IntelPruningCompressor(BaseCompressor, ABC):
    def __init__(self, config_file: str = None):
        super().__init__(config_file)
        self._temp_dir = mkdtemp()

    @property
    def config_key(self) -> str:
        return "intel_pruning"

    @staticmethod
    def _get_default_config() -> Dict:
        # see https://github.com/intel/neural-compressor/blob/master/neural_compressor/conf/config.py  # noqa
        # for further details
        config = {
            "train": {
                "optimizer": {
                    "Adam": {
                        "learning_rate": 0.001,
                        "beta_1": 0.9,
                        "beta_2": 0.999,
                        "epsilon": 1e-07,
                        "amsgrad": False,
                    },
                },
                "criterion": {
                    "SparseCategoricalCrossentropy": {
                        "reduction": "mean",
                        "from_logits": False,
                    },
                },
                "epoch": 10,
                "start_epoch": 0,
                "end_epoch": 10,
                "execution_mode": "eager",  # either eager or graph
                "hostfile": None,  # str for multinode training support
            },
            "approach": {
                "weight_compression": {
                    "initial_sparsity": 0.0,
                    "target_sparsity": 0.97,
                    "start_epoch": 0,
                    "end_epoch": 10,
                },
            },
        }
        return config

    def _prepare_config(self, model: Any):
        pruning_config = copy.deepcopy(self._config)
        config = {
            "model": {
                "name": model.__class__.name,
                "framework": _get_model_framework(model),
            },
            "device": "cpu",
            "tuning": {
                "random_seed": 1978,
                "tensorboard": False,
                "workspace": {"path": self._temp_dir},
            },
            "pruning": pruning_config,
        }
        path_file = Path(self._temp_dir) / "temp.yaml"
        with open(path_file, "w") as f:
            yaml.dump(config, f)
        return path_file

    def compress(
        self,
        model: Any,
        train_input_data: DataManager,
        eval_input_data: DataManager,
        metric_drop_ths: float,
        metric: Callable,
    ) -> Tuple[Any, Optional[float]]:
        config_file = self._prepare_config(model)
        prune = Pruning(config_file)
        prune.model = model
        prune.train_dataloader = self._get_dataloader(train_input_data)
        compressed_model = prune.fit()
        if compressed_model is None:
            return compressed_model, None
        error = self._compute_error(
            model, compressed_model, eval_input_data, metric
        )
        if error > metric_drop_ths:
            return None, None
        perf_loss_ths = metric_drop_ths - error
        return compressed_model, perf_loss_ths

    @abstractmethod
    def _compute_error(
        self,
        model: Any,
        compressed_model: Any,
        eval_input_data: DataManager,
        metric: Callable,
    ):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _get_dataloader(input_data: DataManager):
        raise NotImplementedError


class _IPCDataset(Dataset):
    def __init__(self, input_data: DataManager):
        self._input_data = input_data
        self._internal_size = input_data[0][0][0].shape[0]

    def __getitem__(self, item):
        ptr = item // self._internal_size
        return sum(self._input_data[ptr], ())

    def __len__(self):
        last_el_size = self._input_data[-1][0][0].shape[0]
        return self._internal_size * (len(self._input_data) - 1) + last_el_size


class TorchIntelPruningCompressor(IntelPruningCompressor):
    @staticmethod
    def _get_dataloader(input_data: DataManager):
        bs = input_data[0][0][0].shape[0]
        ds = _IPCDataset(input_data)
        dl = DataLoader(ds, bs)
        return dl

    def _compute_error(
        self,
        model: torch.nn.Module,
        compressed_model: torch.nn.Module,
        eval_input_data: DataManager,
        metric: Callable,
    ):
        if len(eval_input_data) == 0:
            return np.inf
        metric_val = 0
        for inputs, y in eval_input_data:
            pred_model = model(*inputs)
            pred_compressed_model = compressed_model(*inputs)
            metric_val += metric(pred_model, pred_compressed_model, y)
        return metric_val / len(eval_input_data)



from collections import OrderedDict
from typing import (
    Union,
    Iterable,
    List,
    Dict,
    Tuple,
    Type,
    Any,
    Sequence,
    Optional,
)

import numpy as np
import torch

from nebullvm.inference_learners import (
    InferenceLearnerWrapper,
    PytorchBaseInferenceLearner,
    LearnerMetadata,
)

try:
    from transformers import (
        PreTrainedModel,
    )
    from transformers.tokenization_utils import PreTrainedTokenizer
except ImportError:
    # add placeholders for function definition
    PreTrainedModel = None
    PreTrainedTokenizer = None


def _flatten_outputs(
    outputs: Union[torch.Tensor, Iterable]
) -> List[torch.Tensor]:
    new_outputs = []
    for output in outputs:
        if isinstance(output, torch.Tensor):
            new_outputs.append(output)
        else:
            flatten_list = _flatten_outputs(output)
            new_outputs.extend(flatten_list)
    return new_outputs


class _TransformerWrapper(torch.nn.Module):
    """Class for wrappering the Transformers and give them an API compatible
    with nebullvm. The class takes and input of the forward method positional
    arguments and transform them in the input dictionaries needed by
    transformers classes. At the end it also flattens their output.
    """

    def __init__(
        self,
        core_model: torch.nn.Module,
        encoded_input: Dict[str, torch.Tensor],
    ):
        super().__init__()
        self.core_model = core_model
        self.inputs_types = OrderedDict()
        for key, value in encoded_input.items():
            self.inputs_types[key] = value.dtype

    def forward(self, *args: torch.Tensor):
        inputs = {
            key: value for key, value in zip(self.inputs_types.keys(), args)
        }
        outputs = self.core_model(**inputs)
        return tuple(_flatten_outputs(outputs.values()))


def _get_size_recursively(
    tensor_tuple: Union[torch.Tensor, Tuple]
) -> List[int]:
    if isinstance(tensor_tuple[0], torch.Tensor):
        return [len(tensor_tuple)]
    else:
        inner_size = _get_size_recursively(tensor_tuple[0])
        return [len(tensor_tuple), *inner_size]


def _get_output_structure_from_text(
    text: str,
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizer,
    tokenizer_args: Dict,
) -> Tuple[OrderedDict, Type]:
    """Function needed for saving in a dictionary the output structure of the
    transformers model.
    """
    encoded_input = tokenizer([text], **tokenizer_args)
    output = model(**encoded_input)
    structure = OrderedDict()
    for key, value in output.items():
        if isinstance(value, torch.Tensor):
            structure[key] = None
        else:
            size = _get_size_recursively(value)
            structure[key] = size
    return structure, type(output)


def _get_output_structure_from_dict(
    input_example: Dict,
    model: PreTrainedModel,
) -> Tuple[OrderedDict, Type]:
    """Function needed for saving in a dictionary the output structure of the
    transformers model.
    """
    output = model(**input_example)
    structure = OrderedDict()
    for key, value in output.items():
        if isinstance(value, torch.Tensor):
            structure[key] = None
        else:
            size = _get_size_recursively(value)
            structure[key] = size
    return structure, type(output)


def _restructure_output(
    output: Tuple[torch.Tensor],
    structure: OrderedDict,
    output_type: Any = None,
):
    """Restructure the flatter output using the structure dictionary given as
    input.
    """
    output_dict = {}
    idx = 0
    for key, value in structure.items():
        if value is None:
            output_dict[key] = output[idx]
            idx += 1
        else:
            tensor_shape = output[idx].shape[1:]
            output_dict[key] = list(
                torch.reshape(
                    torch.stack(
                        output[idx : int(np.prod(value)) + idx]  # noqa E203
                    ),
                    (*value, *tensor_shape),
                )
            )
            idx += np.prod(value)
    if output_type is not None:
        return output_type(**output_dict)
    return output_dict


class HuggingFaceInferenceLearner(InferenceLearnerWrapper):
    """Class wrapping an InferenceLearner model and giving to it the
    huggingface interface.

    The class fuse both the InterfaceLearner and HuggingFace interfaces, giving
    to the final user a model which can be used whit the prefered API without
    the need of adapting the previous code.

    Attributes:
        network_parameters (ModelParams): Model parameters of the model.
        core_inference_learner (PytorchBaseInferenceLearner): Inference learner
            built using the Pytorch interface.
        output_structure (Dict): Original output structure of the HuggingFace
            model.
        input_names (List[str]): List of all the input keys used for the
            original HuggingFace model.
        output_type (Any, optional): Original output type of the HuggingFace
            model.
    """

    def __init__(
        self,
        core_inference_learner: PytorchBaseInferenceLearner,
        output_structure: OrderedDict,
        input_names: List[str],
        output_type: Any = None,
    ):
        super().__init__(core_inference_learner)
        self.output_structure = output_structure
        self.input_names = input_names
        self.output_type = output_type

    def _save_wrapper_extra_info(self):
        pass

    @staticmethod
    def _load_wrapper_extra_info(builder_inputs: Dict) -> Dict:
        return builder_inputs

    def run(self, *args, **kwargs) -> Any:
        """Run the underlying optimized model for getting a prediction.

        The method has an hybrid interface. It accepts inputs either as
        positional or keyword arguments. If only positional arguments are given
        the method expects the inputs to be in the canonical
        nebullvm interface. If only keyword arguments are given the method
        expects them to be in the HuggingFace interface. Mixed representation
        is not allowed and will result in an error.
        """
        if len(args) > 0 and len(kwargs) > 0:
            raise RuntimeError(
                "Not allowed usage of the predict method. "
                "Either the positional or the keyword arguments must be given."
            )
        if len(args) > 0:
            return self.core_inference_learner(*args)
        inputs = (kwargs.pop(name) for name in self.input_names)
        outputs = self.core_inference_learner(*inputs)
        return _restructure_output(
            outputs, self.output_structure, self.output_type
        )

    def _get_extra_metadata_kwargs(self) -> Dict:
        metadata_kwargs = {
            "output_structure": self.output_structure,
            "output_structure_keys": list(self.output_structure.keys()),
            "input_names": self.input_names,
        }
        if self.output_type is not None:
            metadata_kwargs.update(
                {
                    "output_type": self.output_type.__name__,
                    "output_type_module": self.output_type.__module__,
                }
            )
        return metadata_kwargs

    @staticmethod
    def _convert_metadata_to_inputs(metadata: LearnerMetadata) -> Dict:
        # we need to guarantee the preservation of the output structure
        # elements order.
        output_structure = OrderedDict()
        for key in metadata["output_structure_keys"]:
            output_structure[key] = metadata["output_structure"][key]

        inputs = {
            "output_structure": output_structure,
            "input_names": metadata["input_names"],
        }
        if metadata["output_type"] is not None:
            exec(
                f"from {metadata['output_type_module']} "
                f"import {metadata['output_type']}"
            )
            inputs["output_type"] = eval(metadata["output_type"])
        return inputs


class _HFTextDataset(Sequence):
    def __init__(
        self,
        input_texts: List,
        ys: Optional[List],
        keywords: List[str],
        batch_size: int,
        tokenizer: PreTrainedTokenizer,
        tokenizer_args: Dict,
    ):
        self._input_texts = input_texts
        self._ys = ys
        self._bs = batch_size
        self._keys = keywords
        self._tokenizer = tokenizer
        if self._tokenizer.pad_token is None:
            self._tokenizer.pad_token = self._tokenizer.eos_token
        _tokenizer_args = {"truncation": True, "padding": True}
        _tokenizer_args.update(tokenizer_args)
        self._tokenizer_args = _tokenizer_args

    def __getitem__(self, item: int):
        pointer = self._bs * item
        if pointer >= len(self._input_texts):
            raise IndexError
        mini_batch = self._input_texts[
            pointer : pointer + self._bs  # noqa E203
        ]
        if self._ys is not None:
            mini_batch_y = self._ys[pointer : pointer + self._bs]  # noqa E203
        else:
            mini_batch_y = None
        encoded_inputs = self._tokenizer(mini_batch, **self._tokenizer_args)
        return tuple(encoded_inputs[key] for key in self._keys), mini_batch_y

    def __len__(self):
        return len(self._input_texts) // self._bs


class _HFDictDataset(Sequence):
    def __init__(
        self,
        input_data: List,
        ys: Optional[List],
        keywords: List[str],
    ):
        self._input_data = input_data
        self._ys = ys
        self._keys = keywords

    def __getitem__(self, item: int):
        pointer = item
        if pointer >= len(self._input_data):
            raise IndexError
        mini_batch = self._input_data[pointer]
        if self._ys is not None:
            mini_batch_y = self._ys[pointer]
        else:
            mini_batch_y = None
        return (
            tuple(torch.concat([mini_batch[key]]) for key in self._keys),
            mini_batch_y,
        )

    def __len__(self):
        return len(self._input_data)


def convert_hf_model(
    model: PreTrainedModel,
    input_data: List,
    tokenizer: Optional[PreTrainedTokenizer] = None,
    tokenizer_args: Optional[Dict] = None,
    batch_size: int = 1,
    **kwargs,
):
    if is_dict_type(input_data[0]):
        # already tokenized data
        if "labels" in input_data[0]:
            labels = [data.pop("labels") for data in input_data]
        else:
            labels = None
        input_example = input_data[0]
        output_structure, output_type = _get_output_structure_from_dict(
            input_example=input_example,
            model=model,
        )
        input_data = _HFDictDataset(
            input_data=input_data,
            ys=labels,
            keywords=list(input_example.keys()),
        )

    else:
        assert tokenizer is not None, (
            "Tokenizer is needed when passing data in string format. Please "
            "provide the tokenizer as keyword argument."
        )
        if tokenizer_args is None:
            tokenizer_args = {}
        if not isinstance(input_data[0], str):
            ys = [data[1] for data in input_data]
            input_data = [data[0] for data in input_data]
        else:
            ys = None
        output_structure, output_type = _get_output_structure_from_text(
            text=input_data[0],
            model=model,
            tokenizer=tokenizer,
            tokenizer_args=tokenizer_args,
        )
        input_example = tokenizer(input_data, **tokenizer_args)
        input_data = _HFTextDataset(
            input_texts=input_data,
            ys=ys,
            keywords=list(input_example.keys()),
            batch_size=batch_size,
            tokenizer=tokenizer,
            tokenizer_args=tokenizer_args,
        )
    wrapper_model = _TransformerWrapper(
        core_model=model, encoded_input=input_example
    )
    return (
        wrapper_model,
        input_data,
        list(wrapper_model.inputs_types.keys()),
        output_structure,
        output_type,
    )


def is_dict_type(data_sample: Any):
    try:
        data_sample.items()
    except AttributeError:
        return False
    else:
        return True



import warnings

from nebullvm.installers.installers import install_tf2onnx, install_tensorflow
from nebullvm.utils.general import check_module_version

NoneType = type(None)


class Keras:
    Model = NoneType


class Tensorflow:
    Module = NoneType
    Tensor = NoneType
    keras = Keras()

    @staticmethod
    def function(**kwargs):
        return lambda x: x


try:
    import tensorflow  # noqa F401

    if not check_module_version(tensorflow, min_version="2.7.0"):
        warnings.warn(
            "tensorflow module version must be >= 2.7.0. "
            "Trying to update it."
        )
        install_tensorflow()
        import tensorflow
except ImportError:
    warnings.warn(
        "tensorflow module is not installed on this platform."
        "Please install it if you want to use tensorflow API."
    )
    tensorflow = Tensorflow


try:
    import tf2onnx  # noqa F401
except ImportError:
    try:
        import tensorflow  # noqa F401

        warnings.warn(
            "tf2onnx module is not installed on this platform. "
            "Trying to install it."
        )

        install_tf2onnx()
        import tf2onnx
    except ImportError:
        tf2onnx = object



from typing import Dict, Type

from nebullvm.base import ModelCompiler
from nebullvm.optimizers.base import BaseOptimizer  # noqa F401
from nebullvm.optimizers.blade_disc import BladeDISCOptimizer  # noqa F401
from nebullvm.optimizers.deepsparse import DeepSparseOptimizer  # noqa F401
from nebullvm.optimizers.onnx import ONNXOptimizer  # noqa F401
from nebullvm.optimizers.openvino import OpenVinoOptimizer  # noqa F401
from nebullvm.optimizers.tensor_rt import TensorRTOptimizer  # noqa F401
from nebullvm.optimizers.tvm import ApacheTVMOptimizer  # noqa F401

__all__ = [k for k in globals().keys() if not k.startswith("_")]


COMPILER_TO_OPTIMIZER_MAP: Dict[ModelCompiler, Type[BaseOptimizer]] = {
    ModelCompiler.APACHE_TVM: ApacheTVMOptimizer,
    ModelCompiler.OPENVINO: OpenVinoOptimizer,
    ModelCompiler.TENSOR_RT: TensorRTOptimizer,
    ModelCompiler.ONNX_RUNTIME: ONNXOptimizer,
    ModelCompiler.DEEPSPARSE: DeepSparseOptimizer,
    ModelCompiler.BLADEDISC: BladeDISCOptimizer,
}



from pathlib import Path
from setuptools import setup, find_packages


REQUIREMENTS = [
    "numpy>=1.20.0, <1.23.0",
    "scipy<=1.5.4",
    "onnx>=1.10.0",
    "onnxmltools>=1.11.0",
    "py-cpuinfo>=8.0.0",
    "PyYAML>=6.0",
    "psutil>=5.9.0",
    "requests>=2.28.1",
    "torch>=1.10.0",
    "tqdm>=4.63.0",
    "packaging>=21.3",
]


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="nebullvm",
    version="0.4.3",
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    long_description=long_description,
    include_package_data=True,
    long_description_content_type="text/markdown",
)
